//this file is part of notepad++
//Copyright (C)2003 Don HO ( donho@altern.org )
//
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either
//version 2 of the License, or (at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

// This compiles under Dev-C++ 4.9.9.2 as a .c or .cpp file

// Exclude rarely-used stuff from Windows headers
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
// If you get a [Linker error] undefined reference to `_imp__PathAppendA@8'
// libshlwapi.a will need to be added to the Dev-CPP (Project Options Parameters Add Library)
// /Dev-Cpp/lib/libshlwapi.a
#include <shellapi.h>

/* PathAppend causes a lot of compilers problems. I'd rewrite it out but it seems that compilers,
   particular Microsoft Visual C++ should support it */
#if defined(__MINGW32__) || defined(__BORLANDC__) || defined(__WATCOMC__) || defined(_MSC_VER)
#include <shlwapi.h>
#endif

#if defined(__DMC__) /* Digital Mars */
// WINSHELLAPI isn't the right definition but it turns out that all of these type of definitions are the same so I'll use one that is readily available */
#ifdef __cplusplus
extern "C"
#endif
WINSHELLAPI BOOL APIENTRY PathAppend(LPSTR,LPCSTR);
#endif

#if defined(_MSC_VER) && !defined(PathAppend) /* Visual C++: I can't find shlwapi.h in the SDK so I'll extract the lines from MinGW */

#define WINSHLWAPI DECLSPEC_IMPORT
extern "C" {
WINSHLWAPI BOOL WINAPI PathAppendA(LPSTR,LPCSTR);
WINSHLWAPI BOOL WINAPI PathAppendW(LPWSTR,LPCWSTR);
}
#ifdef UNICODE
#define PathAppend PathAppendW
#else
#define PathAppend PathAppendA
#endif
#endif

#include <shlobj.h>

#include "PluginInterface.h"

#ifndef NELEM
#define NELEM(xyzzy) (sizeof(xyzzy)/sizeof(xyzzy[0]))
#endif
// This is the name which will be displayed in Plugins Menu
#ifdef __cplusplus
#define PLUGIN_NAME "Insert Demo (C++)"
#else
#define PLUGIN_NAME "Insert Demo (C)"
#endif
char iniFilePath[MAX_PATH];
#define sectionName "Insert Extension"
#define keyName "doCloseTag"
unsigned miinsertHtmlCloseTag;

struct NppData nppData;

//----------------------------------------------------------//
// Here're the definition of functions for our functionalities
//----------------------------------------------------------//

#define getCurrentHScintilla(which) ((!which)?nppData._scintillaMainHandle:nppData._scintillaSecondHandle)

void insertCurrentPath(int which) {
	int msg;

  switch (which) {
	case FILE_NAME:
		msg = WM_GET_FILENAME;
    break;
	case CURRENT_DIRECTORY:
		msg = WM_GET_CURRENTDIRECTORY;
    break;
  default:
    msg = WM_GET_FULLCURRENTPATH;
    break;
  }

	char path[MAX_PATH];
	int currentEdit; SendMessage(nppData._nppHandle, WM_GETCURRENTSCINTILLA, 0, (LPARAM)&currentEdit);
	SendMessage(nppData._nppHandle, msg, 0, (LPARAM)path);
	SendMessage(getCurrentHScintilla(currentEdit), SCI_REPLACESEL, 0, (LPARAM)path);
}

PFUNCPLUGINCMD pfinsertCurrentFullPath() {
	insertCurrentPath(FULL_CURRENT_PATH);
}
PFUNCPLUGINCMD pfinsertCurrentFileName() {
	insertCurrentPath(FILE_NAME);
}
PFUNCPLUGINCMD pfinsertCurrentDirectory() {
	insertCurrentPath(CURRENT_DIRECTORY);
}

void insertDateTime(DWORD format) {
	char date[128],time[128],dateTime[256];
	SYSTEMTIME st;

  GetLocalTime(&st);
	GetDateFormat(LOCALE_USER_DEFAULT, format, &st, NULL, date, sizeof(date));
	GetTimeFormat(LOCALE_USER_DEFAULT, TIME_NOSECONDS, &st, NULL, time, sizeof(time));

  wsprintf(dateTime, "%s %s", time, date);

	int currentEdit; SendMessage(nppData._nppHandle, WM_GETCURRENTSCINTILLA, 0, (LPARAM)&currentEdit);
	SendMessage(getCurrentHScintilla(currentEdit), SCI_REPLACESEL, 0, (LPARAM)dateTime);
}

PFUNCPLUGINCMD pfinsertShortDateTime() {
	insertDateTime(DATE_SHORTDATE);
}

PFUNCPLUGINCMD pfinsertLongDateTime() {
	insertDateTime(DATE_LONGDATE);
}

PFUNCPLUGINCMD pfinsertHtmlCloseTag() ;

PFUNCPLUGINCMD pfabout() {
	MessageBox(nppData._nppHandle,
   "Are there things you want in an editor but can't find? This\r\n"
   "Notepad++ Plugin is a DLL written in C or C++. It doesn't do\r\n"
   "much but you can add to it to give the editor the functionality\r\n"
   "you want. The editor and plugin source code are available at\r\n"
   "http://notepad-plus.sourceforge.net/\r\n"
   "with which you can make your own plugins.\r\n"
   "If you're a programmer, you'll discover that your favorite editor\r\n"
   "is not the one with the most features, it's the one you can write a\r\n"
   "plugin to add your own features."
   , PLUGIN_NAME, MB_OK);
}

struct FuncItem funcItem[]={
{"Current Full Path",pfinsertCurrentFullPath,0,FALSE},
{"Current File Name",pfinsertCurrentFileName,0,FALSE},
{"Current Directory",pfinsertCurrentDirectory,0,FALSE},
{"Date && Time - short format",pfinsertShortDateTime,0,FALSE},
{"Date && Time - long format",pfinsertLongDateTime,0,FALSE},
{"Close HTML/XML tag automatically",pfinsertHtmlCloseTag,0,FALSE},
{"About",pfabout,0,FALSE},
};

PFUNCPLUGINCMD pfinsertHtmlCloseTag() {
  funcItem[miinsertHtmlCloseTag]._init2Check = !funcItem[miinsertHtmlCloseTag]._init2Check;
	CheckMenuItem(GetMenu(nppData._nppHandle), funcItem[miinsertHtmlCloseTag]._cmdID, MF_BYCOMMAND | (funcItem[miinsertHtmlCloseTag]._init2Check?MF_CHECKED:MF_UNCHECKED));
}

unsigned findmenuitem(PFUNCPLUGINCMD_MSC *_pFunc) {
  unsigned itemno; for(itemno=0; itemno<NELEM(funcItem); itemno++) if (_pFunc == funcItem[itemno]._pFunc) return(itemno);
  return(0);
}

LRESULT DLLIMPORT __cdecl messageProc(UINT Message, WPARAM wParam, LPARAM lParam) {
  return(0);
}

/*
 *--------------------------------------------------
 * The 4 extern functions are mandatory
 * They will be called by Notepad++ plugins system
 *--------------------------------------------------
*/

// The setInfo function gets the needed infos from Notepad++ plugins system
void DLLIMPORT __cdecl setInfo(struct NppData notpadPlusData) {
	nppData = notpadPlusData;
}

// The getName function tells Notepad++ plugins system its name
const char DLLIMPORT * __cdecl getName(void) {
	return PLUGIN_NAME;
}

// The getFuncsArray function gives Notepad++ plugins system the pointer FuncItem Array
// and the size of this array (the number of functions)
struct FuncItem DLLIMPORT * __cdecl getFuncsArray(int *nbF) {
	*nbF = NELEM(funcItem);
	return funcItem;
}

// If you don't need get the notification from Notepad++,
// just let it be empty.
void DLLIMPORT __cdecl beNotified(struct SCNotification *notifyCode) {
  enum LangType docType;

	switch (notifyCode->nmhdr.code) {
  case SCN_CHARADDED:
    SendMessage(nppData._nppHandle, WM_GETCURRENTLANGTYPE, 0, (LPARAM)&docType);
    BOOL isDocTypeHTML = (docType == L_HTML || docType == L_XML || docType == L_PHP);
    if (funcItem[miinsertHtmlCloseTag]._init2Check && isDocTypeHTML) {
      if (notifyCode->ch == '>') {
        char buf[512];
        int currentEdit;
        SendMessage(nppData._nppHandle, WM_GETCURRENTSCINTILLA, 0, (LPARAM)&currentEdit);
        HWND hCurrentEditView = getCurrentHScintilla(currentEdit);
        int currentPos = (int)(SendMessage(hCurrentEditView, SCI_GETCURRENTPOS, 0, 0));
        int beginPos = currentPos - (sizeof(buf) - 1);
        int startPos = (beginPos > 0)?beginPos:0;
        int size = currentPos - startPos;
        int insertStringSize = 2;
        char insertString[516] = "</";

        if (size >= 3) {
          struct TextRange tr = {{startPos, currentPos}, buf};

          SendMessage(hCurrentEditView, SCI_GETTEXTRANGE, 0, (LPARAM)&tr);

          if (buf[size-2] != '/') {

            const char* pBegin = &buf[0];
            const char* pCur = &buf[size - 2];
            int  insertStringSize = 2;

            for (; pCur > pBegin && *pCur != '<' && *pCur != '>' ;)
              pCur--;

            if (*pCur == '<') {
              pCur++;

              while (strchr(":_-.", *pCur) || IsCharAlphaNumeric(*pCur)) {
                insertString[insertStringSize++] = *pCur;
                pCur++;
              }
            }

            insertString[insertStringSize++] = '>';
            insertString[insertStringSize] = '\0';

            if (insertStringSize > 3) {
              SendMessage(hCurrentEditView, SCI_BEGINUNDOACTION, 0, 0);
              SendMessage(hCurrentEditView, SCI_REPLACESEL, 0, (LPARAM)insertString);
              SendMessage(hCurrentEditView, SCI_SETSEL, currentPos, currentPos);
              SendMessage(hCurrentEditView, SCI_ENDUNDOACTION, 0, 0);
            }
          }
        }
      }
    }
    break;
	}
}

//----------------------------------------------------------//
// Standard Windows DLL Entry Point
//----------------------------------------------------------//

BOOL APIENTRY DllMain (HINSTANCE hInst     /* Library instance handle. */ ,
                       DWORD reason        /* Reason this function is being called. */ ,
                       LPVOID reserved     /* Not used. */ ) {
  ITEMIDLIST *pidl;

  switch (reason) {
  case DLL_PROCESS_ATTACH:
    miinsertHtmlCloseTag=findmenuitem(pfinsertHtmlCloseTag);
    SHGetSpecialFolderLocation(NULL, CSIDL_APPDATA, &pidl);
    SHGetPathFromIDList(pidl, iniFilePath);
    PathAppend(iniFilePath, "Notepad++\\insertExt.ini");

    funcItem[miinsertHtmlCloseTag]._init2Check = (GetPrivateProfileInt(sectionName, keyName, 1, iniFilePath) != 0);
    break;
  case DLL_PROCESS_DETACH:
    WritePrivateProfileString(sectionName, keyName, funcItem[miinsertHtmlCloseTag]._init2Check?"1":"0", iniFilePath);
    break;
  case DLL_THREAD_ATTACH:
    break;
  case DLL_THREAD_DETACH:
    break;
  }
  return TRUE;
}
/* Execution sequence:
  1) DllMain is called when Notepad++ loads the plugin
  2) N++ calls getFuncsArray, setInfo, getName to get initial operating info
  3) beNotified is called constantly while editing.
  4) Occasionally the user will select a plugin menu item.
  5) The user will quit N++ which will call DllMain for the last time. */
